"""
Defaults
--------

.. data:: SERVER_LISTEN
  :noindex:

  Configuration default listen address (``localhost``)

.. data:: SERVER_PORT
  :noindex:

  Configuration default listen port (``8080``)

.. data:: CFGFILE_SEARCH_PATHS
  :noindex:

  Configuration default search paths

  .. code-block:: python

    [
        ".<filename>.yml",
        "<filename>.yml",
        "$XDG_CONFIG_HOME/<filename>.yml",
        "$HOME/.<filename>.yml",
        "/etc/xdg/<filename>.yml",
        "/etc/<filename>.yml",
    ]

Default configuration
---------------------

.. code-block:: python

  {
      # The following can be set in the configuration file:
      "server": {
          "debug": False,
          "listen": SERVER_LISTEN,
          "port": SERVER_PORT,
          "ui": True,
          "workers": 0,
          "pidfile": None,
          "proxy": {
              "x_for": 0,
              "x_proto": 0,
              "x_host": 0,
              "x_port": 0,
              "x_prefix": 0,
          },
          "cachedir": None,
          "timeout": 120,
      },
      "celery": {
          "broker_url": "redis://localhost:6379/0",
          "result_backend": "redis://localhost:6379/0",
          "broker_connection_retry_on_startup": False,
      },
  }

"""
import os
import typing as t

from boltons.iterutils import flatten_iter, unique_iter
from xdg import XDG_CONFIG_DIRS

SERVER_LISTEN: str = "localhost"
SERVER_PORT: int = 8080

# fmt: off
CFGFILE_SEARCH_PATHS: t.List[str] = list(
    unique_iter(
        flatten_iter(
        [
            "./.{cfgfilename}",                # ./.<filename>.yml
            "./{cfgfilename}",                 # ./<filename>.yml
            "$XDG_CONFIG_HOME/{cfgfilename}",  # ~/.config/<filename>.yml
            "~/.{cfgfilename}",                # ~/.<filename>.yml
            [os.path.join(str(e), "{cfgfilename}") for e in XDG_CONFIG_DIRS],
            "/etc/{cfgfilename}",              # /etc/<filename>.yml
        ],
        ),
    ),
)
# fmt: on


def make_config(defaults: t.Dict[str, t.Any] = None) -> t.Dict[str, t.Any]:
    """
    Convenience function that returns defaults for arguments,
    and for configuration file.

    Parameters
    ----------
    defaults: dict
        a dictionary containing default values.

    Returns
    -------
    dict:
        a dictionary
    """
    defaults = defaults or {}
    defaults["server"] = defaults.get("server") or {}

    # cf. src/middleware/__init__.py
    return {
        "server": {
            "debug": defaults["server"].get("debug", False),
            "listen": defaults["server"].get("listen", SERVER_LISTEN),
            "port": defaults["server"].get("port", SERVER_PORT),
            "ui": defaults["server"].get("ui", True),
            "workers": defaults["server"].get("workers", 0),
            "pidfile": defaults["server"].get("pidfile", None),
            "proxy": {
                "x_for": defaults["server"].get("proxy", {}).get("x_for", 0),
                "x_proto": defaults["server"].get("proxy", {}).get("x_proto", 0),
                "x_host": defaults["server"].get("proxy", {}).get("x_host", 0),
                "x_port": defaults["server"].get("proxy", {}).get("x_port", 0),
                "x_prefix": defaults["server"].get("proxy", {}).get("x_prefix", 0),
            },
            "cachedir": None,
            "timeout": 120,
        },
        "celery": {
            "broker_url": "redis://localhost:6379/0",
            "result_backend": "redis://localhost:6379/0",
            "broker_connection_retry_on_startup": False,
        },
    }



import contextlib
import logging

from flask import current_app
from werkzeug.local import LocalProxy

with contextlib.suppress(ImportError):
    from celery.utils.log import get_task_logger


def __get_current_logger():
    # Without a context, the stderr logger is still valid
    if not current_app:
        return logging.getLogger("stderr")

    # If we're in celery, return a celery logger
    celery_app = current_app.extensions.get("celery")
    if celery_app and celery_app.current_worker_task:
        return get_task_logger(__name__)

    # Users shouldn't directly log requests, so return
    # the stderr logger
    return logging.getLogger("stderr")


def _get_current_logger():
    return __get_current_logger()


# pylint: disable=unnecessary-lambda
current_logger = LocalProxy(lambda: _get_current_logger())



"""
Description
-----------

Options
-------

.. program:: flasket-dev

.. option:: -l HOST, --listen HOST

   The ip to listen on (default: localhost)

.. option:: -p PORT, --port PORT

   The port to listen on (default: 8080)

.. option:: -c CFGFILE, --cfgfile CFGFILE

   Use CFGFILE as configuration file, otherwise first file found in
   search path is used. (default search path: (...))

.. option:: --ui, --no-ui

   Enable the OpenAPI UI. Disable with :option:`--no-ui`. (default: enabled)

.. option:: --debug, --no-debug

  Enable debug mode. Disable with :option:`--no-debug`. (default: disabled)

Usage
-----

.. program:: flasket-dev

.. code-block:: python

  #!/usr/bin/env python3

  from flasket.cli import make_flasket

  if __name__ == "__main__":
      flasket = make_flasket_cli(__file__, "flask", cfgname="example.yml")
      flasket.run()


"""
import argparse
import typing as t


class FlaskHelper:
    @staticmethod
    def force_cfg(cfg: t.Dict["str", t.Any]) -> None:
        # debug is set via flag
        cfg["server"]["production"] = False
        return cfg

    @staticmethod
    def add_arguments(defaults: t.Dict["str", t.Any], parser: argparse.ArgumentParser) -> None:
        """
        Adds Flask middleware specific command line arguments :option:`--debug`, :option:`--no-debug`
        to the ArgumentParser.

        Parameters
        ----------
        defaults:

            a dictionnary containing default values

        parser: argparse.ArgumentParser

            a parser on which `add_argument` will be called
        """
        debug = {True: "enabled", False: "disabled"}[defaults["debug"]]

        # fmt: off
        parser.add_argument(
            "--debug", action="store_true", default=None,
            help=f"Enable debug mode. Disable with --no-debug. (default: {debug})",
        )
        parser.add_argument(
            "--no-debug", action="store_false", default=None, dest="debug",
            help=argparse.SUPPRESS,
        )
        # fmt: on



"""
Description
===========

Parse command line arguments for Flask, or for Celery, and export
the contents of the yaml configuration file passed by `--cfgfile` (Flask)
or `--conf` (Celery) to the Flask app.

No settings are stored directly in Celery app.

When ArgParseExt is used on the Flask application, it will parse application
arguments and display the help if needed.

When used on the Celery application, it will parse application arguments
only to find the `--conf` flag, equivalent to the Flask `--cfgfile` argument.

ArgParseExt then searchs for the configuration file according to it's name,
"settings.yml" by default. The default search path will be shown in the `--help`
command printout.

`default_cfg` and `forced_cfg` are relative to the root contents of the settings
file. In other words, when passing Celery configuration, the celery settings must
be under the `celery` key.


.. docs/ext/argparse/flask.rst
.. include:: argparse/flask.rst

.. docs/ext/argparse/gunicorn.rst
.. include:: argparse/gunicorn.rst


Usage
=====

.. program:: flask-dev

Initializing applications
-------------------------

.. code-block:: python

  #!/usr/bin/env python3

  from flask import Flask
  from flask_gordon.ext import ArgParseExt
  from flask_gordon.ext import CeleryExt

  # Initialize Flask
  app = Flask(__name__)

  # Initialize Celery
  celery = CeleryExt().init_app(app)
  celery = ArgParseExt().init_app(
      forced_cfg={
          "imports": "<PACKAGE_NAME>.tasks",
      },
  )

  # If this is Flask:
  if __name__ == "__main__":
      global app
      app = ArgParseExt().init_app(app)
      # This line is reached if parsing of arguments did
      # not raise an error and was not `--help`.


Classes
=======

.. autoclass:: ArgParseExt
   :members: __init__, init_app

"""
import argparse
import os
import sys
import typing as t

from torxtools.argtools import is_not_dir

from .. import functions
from ._functions import deepmerge, prepare_configuration, read_configuration
from .flask import FlaskHelper
from .gunicorn import GunicornHelper


class ArgParseExt:
    def __init__(
        self,
        app=None,
        middleware: str = "gunicorn",
        argv: t.List[str] = None,
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        forced_cfg: t.Dict[str, t.Any] = None,
        description="Flasket server",
        appname: str = None,
    ):
        """
        Parameters
        ----------
        app:

            A Flask or Celery application.

        middleware: str

            Middleware to use: 'gunicorn' for production, 'flask' for debug.
            Unused in the case of a Celery app.

        argv: list[str]

            Uses :code:`sys.argv[1:]` if None. Use :code:`[]` if you desire to deactive argument parsing.

        cfgfilename: str

            Configuration filename to use. Typically this is your packagename with a 'yml' extension.

        cfgfilepaths: str

            Paths to search for in order to find the configuration file. Every value of the list must contain the placeholder '{cfgfilename}' such as :code:`["/home/jdoe/{cfgfilename}"]`

        default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

            Dictionary containing the defaults for the command line arguments and configuration.
            Passed value will be merged with the default factory configuration, command line arguments
            and the configuration file that was read.

        forced_cfg: dict, default: {}

            Dictionary containing the forced settings.
            Passed value will be merged with the default factory configuration, command line arguments
            and the configuration file that was read, and then the forced_cfg will overwrite the values.
            Flask middleware forced configuration will overwrite the final configuration.

        description: str

            Text description to use in argparse.
            Unused in the case of a Celery app.

        appname: str

            Application name. Used to show application name in `--help` string.
            Defaults to sys.argv[0]
            Unused in the case of a Celery app.

        """
        if app is not None:
            self.init_app(
                app,
                middleware,
                argv=argv,
                cfgfilename=cfgfilename,
                cfgfilepaths=cfgfilepaths,
                default_cfg=default_cfg,
                forced_cfg=forced_cfg,
                description=description,
                appname=appname,
            )

    def init_app(
        self,
        app,
        middleware: str = "gunicorn",
        argv: t.List[str] = None,
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        forced_cfg: t.Dict[str, t.Any] = None,
        description="Flask server",
        appname: str = None,
    ):
        # Is this a flask or celery application?
        app_class = functions.get_app_class(app)
        if app_class == "flask":
            return self.init_flask_app(
                app,
                middleware,
                argv=argv,
                cfgfilename=cfgfilename,
                cfgfilepaths=cfgfilepaths,
                default_cfg=default_cfg,
                forced_cfg=forced_cfg,
                description=description,
                appname=appname,
            )

        if app_class == "celery":
            return self.init_celery_app(
                app,
                argv=argv,
                cfgfilename=cfgfilename,
                cfgfilepaths=cfgfilepaths,
                default_cfg=default_cfg,
                forced_cfg=forced_cfg,
            )

        raise NotImplementedError("Could not determine if app is Flask or Celery")

    def init_celery_app(
        self,
        app: "Celery",
        argv: t.List[str] = None,
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        forced_cfg: t.Dict[str, t.Any] = None,
    ):
        celery_app = app

        cfgfilename, cfgfilepaths, defaults = prepare_configuration(cfgfilename, cfgfilepaths, default_cfg)

        # Parse arguments if they exist
        if argv is None:
            argv = sys.argv[1:]

        cfgfile = None
        if "--config" in argv:
            idx = argv.index("--config")
            cfgfile = argv[idx + 1]

        cfgfileread, data = read_configuration(cfgfile, cfgfilename, cfgfilepaths)

        # Merge in the reverse order of priority
        cfg = defaults
        cfg = deepmerge(cfg, data)
        cfg = deepmerge(cfg, forced_cfg or {})

        if "celery" not in cfg:
            cfg["celery"] = {}

        # Copy the configuration to the app.
        # When celery starts, the ArgParseExt of app is not called since argv differs
        # between starting the Flask application and the Celery application
        #
        if getattr(celery_app, "app", None):
            celery_app.app().config["SETTINGS"] = cfg

        # Create a celery application
        celery_app.config_from_object(cfg["celery"])
        celery_app.set_default()

        # Save the configuration file used on the celery app, we'll print it if
        # the user also uses LoggerExt
        # pylint: disable=protected-access
        celery_app._cfgfileread = cfgfileread

        return celery_app

    def init_flask_app(
        self,
        app: "Flask",
        middleware: str = "gunicorn",
        argv: t.List[str] = None,
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        forced_cfg: t.Dict[str, t.Any] = None,
        description="Flask server",
        appname: str = None,
    ):
        flask_app = app

        # Verify middleware exists
        middleware = middleware.lower().strip()
        if middleware not in ["flask", "gunicorn"]:
            raise ValueError('middleware argument must be in ["flask", "gunicorn"]')
        if middleware == "flask":
            helper = FlaskHelper()
        if middleware == "gunicorn":
            helper = GunicornHelper()

        # Parse arguments if they exist
        if argv is None:
            argv = sys.argv[1:]
            if appname is None:
                appname = os.path.basename(sys.argv[0])

        cfgfilename, cfgfilepaths, defaults = prepare_configuration(cfgfilename, cfgfilepaths, default_cfg)

        arguments = self._parse_arguments(
            argv=argv,
            cfgfilepaths=cfgfilepaths,
            helper=helper,
            defaults=defaults,
            description=description,
        )

        # Copy arguments over to an dict with server
        arguments = {k: v for k, v in arguments.items() if v is not None}
        cfgfile = arguments.pop("cfgfile", None)
        arguments = {"server": arguments}

        _, data = read_configuration(cfgfile, cfgfilename, cfgfilepaths)

        # Merge in the reverse order of priority
        cfg = defaults
        cfg = deepmerge(cfg, data)
        cfg = deepmerge(cfg, arguments)
        cfg = deepmerge(cfg, forced_cfg or {})
        cfg = helper.force_cfg(cfg)

        flask_app.config["SETTINGS"] = cfg
        return flask_app

    @staticmethod
    def _parse_arguments(*, argv, cfgfilepaths, helper, defaults, description):
        if not argv:
            return {}

        # argument_default=None does not set the default to None for boolean options,
        # so we'll specifically set default=None for those values
        #
        # Default values aren't actually added/set here, but in the FlasketSettings,
        # We only care about values that were specified.
        parser = argparse.ArgumentParser(
            description=description,
            argument_default=None,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        # Create helptext for UI option
        b_ui = {True: "enabled", False: "disabled"}[defaults["server"]["ui"]]
        if cfgfilepaths:
            # Keep on two lines, otherwise line continuation will make
            # an obsure argparse bug appear
            helpmsg_cfgfile = "Use CFGFILE as configuration file, "
            helpmsg_cfgfile += (
                f"otherwise first file found in search path is used. (default search path: {cfgfilepaths})"
            )
        else:
            helpmsg_cfgfile = "Use CFGFILE as configuration file."

        # fmt: off
        parser.add_argument(
            "-l", "--listen", metavar="HOST",
            help=f'The ip to listen on (default: {defaults["server"]["listen"]})',
        )
        parser.add_argument(
            "-p", "--port", metavar="PORT", type=int,
            help=f'The port to listen on (default: {defaults["server"]["port"]})',
        )
        parser.add_argument(
         "-c", "--cfgfile", metavar="CFGFILE",
           help=helpmsg_cfgfile,
           type=is_not_dir,
        )
        parser.add_argument(
            "--ui", action="store_true", default=None,
            help=f"Enable the OpenAPI UI. Disable with --no-ui. (default: {b_ui})",
        )
        parser.add_argument(
            "--no-ui", action="store_false", default=None, dest="ui",
            help=argparse.SUPPRESS,
        )
        # fmt: on
        helper.add_arguments(defaults["server"], parser)
        args = parser.parse_args(argv)
        return vars(args)



import contextlib
import sys


def get_app_class(app):
    # Is this a flask or celery application?
    with contextlib.suppress(AttributeError):
        _ = app.run
        return "flask"

    with contextlib.suppress(AttributeError):
        _ = app.control.inspect
        return "celery"

    return None


def is_celery_app():
    return "celery.bin.celery" in sys.modules


def is_celery_worker_app():
    if not is_celery_app():
        return False
    return "worker" in sys.argv[1:]



"""
Description
===========

Initialize loggers for Celery workers.

Usage
=====

Initializing a Celery application
---------------------------------

Sample :file:`<PACKAGE_NAME>/__init__.py` file:

.. code-block:: python

  #!/usr/bin/env python3

  from flask import Flask
  from flask_gordon.ext import CeleryExt

  app = Flask(__name__)
  celery = CeleryExt().init_app(app)
  celery = LoggerExt().init_app(celery)

Classes
=======

.. autoclass:: CeleryExt
   :members: init_app

"""
import logging

from .. import functions


class LoggerExt:
    def __init__(self, app=None):
        """
        Parameters
        ----------
        app:

            A Flask or Celery application.
        """
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """
        Parameters
        ----------
        app:

            A Flask or Celery application.
        """
        # Is this a flask or celery application?
        app_class = functions.get_app_class(app)
        if app_class == "flask":
            return self.init_flask_app(app)

        if app_class == "celery":
            return self.init_celery_app(app)

        raise NotImplementedError("Could not determine if app is Flask or Celery")

    def init_flask_app(self, app):
        return app

    def init_celery_app(self, app):
        # Defaults:
        #   https://docs.celeryq.dev/en/stable/userguide/configuration.html#std-setting-worker_log_format
        #
        # We don't need to toy around with the loggers, so we can just change the settings
        app.conf.update(
            {
                "worker_log_format": "[%(asctime)s] [%(levelname)s] [%(processName)s] %(message)s",
                "worker_task_log_format": "[%(asctime)s] [%(levelname)s] [%(processName)s] [%(task_name)s] [%(task_id)s] %(message)s",
                "worker_redirect_stdouts_level": "DEBUG",
            },
        )

        # Print configuration file only in Celery worker application
        # pylint: disable=protected-access
        if functions.is_celery_worker_app() and getattr(app, "_cfgfileread"):
            cfg = app._cfgfileread
            if cfg:
                logging.getLogger("stderr").info(f"Read Celery configuration from file '{cfg}'")
        return app



"""
Description
===========

This middleware creates a single WSGI application that dispatches to
multiple other WSGI applications mounted at different URL paths.

A common example is writing a Single Page Application, where you have a
backend API and a frontend written in JavaScript that does the routing
in the browser rather than requesting different pages from the server.
The frontend is a single HTML and JS file that should be served for any
path besides "/api".

This example dispatches to an API app under "/api", an admin app
under "/admin", and an app that serves frontend files for all other
requests::

    app = Dispatcher(serve_frontend, {
        '/api': api_app,
        '/admin': admin_app,
    })

In production, you might instead handle this at the HTTP server level,
serving files or proxying to application servers based on location. The
API and admin apps would each be deployed with a separate WSGI server,
and the static files would be served directly by the HTTP server.

Unlike the :class:`werkzeug.middleware.Dispatcher` a call to
`/api` will not be routed to the app handling the prefix `/api`.
A call to `/api/` will though.


Usage
=====

.. code-block:: python

    from flask import Flask
    from flask_gordon.middleware import Dispatcher

    paths = {
        "/api": Flask("api", static_folder="api"),
        "/app": Flask("app", static_folder="app"),
    }

    app = Flask(__name__)
    app.wsgi_app = Dispatcher(app.wsgi_app, paths)


Classes
=======

.. autoclass:: Dispatcher

"""
import typing as t

if t.TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment

__all__: t.List[str] = []


class Dispatcher:
    """
    Combine multiple applications as a single WSGI application.
    Requests are dispatched to an application based on the path it is
    mounted under.

    Parameters
    ----------

    app: WSGIApplication

        The WSGI application to dispatch to if the request
        doesn't match a mounted path.

    mounts: t.Dict[str, "WSGIApplication"]

        Maps path prefixes to applications for dispatching.

    """

    def __init__(
        self,
        app: "WSGIApplication",
        mounts: t.Optional[t.Dict[str, "WSGIApplication"]] = None,
    ) -> None:
        self.app = app
        self.mounts = mounts or {}

    def __call__(self, environ: "WSGIEnvironment", start_response: "StartResponse") -> t.Iterable[bytes]:
        script = environ.get("PATH_INFO", "")

        while "/" in script:
            script, _ = script.rsplit("/", 1)
            path = script + "/"

            for mount, app in self.mounts.items():
                prefix = "/" if mount == "/" else mount + "/"
                if prefix == path:
                    return app(environ, start_response)

        return self.app(environ, start_response)



import typing as t
from copy import deepcopy

from boltons.iterutils import remap
from torxtools import xdgtools
from torxtools.cfgtools import which
from torxtools.pathtools import expandpath
from yaml import safe_load

from ..defaults import CFGFILE_SEARCH_PATHS, make_config


def _strip_none(_p: str, k: str, v: t.Any) -> bool:
    return k is not None and v is not None


def _deepmerge(src: t.Dict[str, t.Any], dst: t.Dict[str, t.Any], path: t.List[str] = None) -> t.Dict[str, t.Any]:
    """
    Take every key, value from src and merge it recursively into dst.
    Adapted from https://stackoverflow.com/questions/7204805

    Parameters
    ----------
    dst: dict
        Destination dictionary

    src: dict
        Source dictionary

    path:
        Used internally

    Returns
    -------
    dict:
        Merged dictionary
    """
    if path is None:
        path = []

    for key in src:
        if key not in dst:
            dst[key] = src[key]
        elif isinstance(dst[key], dict) and isinstance(src[key], dict):
            _deepmerge(src[key], dst[key], path + [str(key)])
        elif dst[key] == src[key]:
            pass  # same leaf value
        else:
            dst[key] = src[key]
    return dst


def deepmerge(dst: t.Dict[str, t.Any], src: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """
    Take every key, value from src and merge it recursivly into dst.
    None values are stripped from src before merging.

    Adapted from https://stackoverflow.com/questions/7204805

    Parameters
    ----------
    dst: dict
        destination dictionary

    src: dict
        source dictionary

    Returns
    -------
    dict:
        merged dictionary
    """
    # pylint: disable=unnecessary-lambda-assignment
    src = remap(src, visit=_strip_none)

    return _deepmerge(src, deepcopy(dst))


def _read(cfgfile: str) -> t.Dict[str, t.Any]:
    """
    Convenience function in order to be mocked.

    Parameters
    ----------
    cfgfile: str

        a single path representing a yaml file.

    Returns
    -------
    dict:

        a dictionary
    """
    with open(cfgfile, encoding="UTF-8") as fd:
        data = safe_load(fd)
    return data or {}


def prepare_configuration(cfgfilename, cfgfilepaths, default_cfg):
    # Sets environment variables for XDG paths
    xdgtools.setenv()

    # Prepare search path for configuration file. Disable it if we're looking
    # for a impossible (None) file name
    cfgfilename, cfgfilepaths = _prepare_search_paths(cfgfilename, cfgfilepaths)

    # Create a default configuration from what was passed
    # and what we set. Other values are filtered
    defaults = make_config(default_cfg)
    return cfgfilename, cfgfilepaths, defaults


def _prepare_search_paths(cfgfilename, cfgfilepaths):
    # Prepare search path for configuration file. Disable it if we're looking
    # for a impossible (None) file name
    if not cfgfilename:
        cfgfilename = "settings.yml"
    if cfgfilepaths is None:
        cfgfilepaths = CFGFILE_SEARCH_PATHS
    cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]
    return cfgfilename, cfgfilepaths


def read_configuration(cfgfile, cfgfilename, cfgfilepaths):
    # Search for the configuration file
    if cfgfile is None and cfgfilepaths:
        # Search for cfgfile
        cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]
        cfgfile = which(cfgfile, expandpath(cfgfilepaths))

    return cfgfile, _read(cfgfile or "/dev/null") or {}



from contextlib import contextmanager

from .argparse import ArgParseExt
from .ctx import current_logger
from .logger import LoggerExt
from .testing import FlaskTestingExt

__all__ = [
    # Extensions
    "ArgParseExt",
    "FlaskTestingExt",
    "LoggerExt",
    # LocalProxies
    "current_logger",
]


@contextmanager
def _import_if_possible():
    try:
        yield
    # pylint: disable=bare-except
    except:
        pass


with _import_if_possible():
    from .celery import CeleryExt

    __all__ += ["CeleryExt"]



"""
Description
===========

Initialize a Celery worker or Celery beat for the Flask application.

A weak reference to the Flask application will be added the Celery application.


Usage
=====

Initializing a Celery application
---------------------------------

Sample :file:`<PACKAGE_NAME>/__init__.py` file:

.. code-block:: python

  #!/usr/bin/env python3

  from flask import Flask
  from flask_gordon.ext import CeleryExt

  app = Flask(__name__)
  celery = CeleryExt().init_app(app)

  # Assert references in both directions:
  assert app is celery.app()
  assert app.extensions["celery"] is celery


Sample Celery shared task
-------------------------

.. code-block:: python

  from celery import shared_task
  from flask import current_app

  def call_example(*args, **kwargs):
      example_task.s(*args, **kwargs).apply_async()

  @shared_task()
  def example_task(*args, **kwargs):
      print(current_app)


Starting a Celery worker
-------------------------

Sample :file:`bin/celery-worker` file:

.. code-block:: bash

 #!/usr/bin/env bash

 exec celery -A <PACKAGE_NAME>.celery worker

Classes
=======

.. autoclass:: CeleryExt
   :members: init_app

"""
import contextlib
import typing as t
import weakref
from datetime import datetime

HAS_CELERY = False
with contextlib.suppress(ImportError):
    from celery import Celery, Task

    HAS_CELERY = True


# https://github.com/celery/celery/issues/4400#issuecomment-393390339
class LocalCelery(Celery):
    def now(self):
        """
        Return the current time and date as a datetime.
        """
        return datetime.now(self.timezone)


class CeleryExt:
    def __init__(self):
        if not HAS_CELERY:
            raise NotImplementedError("CeleryExt requires celery or celery[redis] package")

    def init_app(self, app: "Flask", **kwargs):
        """
        Parameters
        ----------
        app: FlaskApp

            A Flask application.

        kwargs:

            Keyword named arguments.

        Returns
        -------
        object:
            A Celery application
        """

        # pylint: disable=abstract-method
        class FlaskTask(Task):
            def __call__(self, *args: object, **kwargs: t.Dict[str, t.Any]) -> object:
                with app.app_context():
                    return self.run(*args, **kwargs)

        if "task_cls" not in kwargs:
            kwargs["task_cls"] = FlaskTask

        # pylint: disable=attribute-defined-outside-init
        celery_app = LocalCelery(app.name, **kwargs)
        celery_app.set_default()

        # Add a weak reference on celery_app to the Flask application
        celery_app.app = weakref.ref(app)

        app.extensions["celery"] = celery_app

        return celery_app
